# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import enum
import binascii

from . import NoData, MissingImport
from .type import *

__all__ = ["NoData", "MissingImport", "Type", "String", "Path", "Number", "FileType", "Mode",
        "Size", "Time", "Duration", "Boolean", "ListOfStrings", "Plugin", "Speed"]


class Speed(enum.Enum):
    """The speed at which a plugin extracts information from an entry. This is
       used as a hint to group faster tests before slower tests in test
       sequences.
    """
    FAST = 0
    NORMAL = 1
    SLOW = 2


class Plugin:
    """Base class for all plugins.
    """

    author = "Lars Gustäbel"
    email = "lars@gustaebel.de"
    url = "https://github.com/gustaebel/ff/"

    use_cache = False
    cache_tag = None
    speed = Speed.NORMAL
    attributes = {}

    def __init_subclass__(cls):
        super().__init_subclass__()

        cls.name = cls.__name__.lower()

    @classmethod
    def initialize(cls, logger, source, path):
        """Initialize the Plugin class with basic information.
        """
        cls.logger = logger
        cls.source = source
        cls.path = path
        if cls.cache_tag is None:
            cls.cache_tag = cls.get_plugin_cache_tag(path)
        assert isinstance(cls.cache_tag, int) and cls.cache_tag >= 0
        cls.sql_table_name = f"plugin_{cls.name}_{cls.cache_tag}"

    def __init__(self):
        self.setup()

    @classmethod
    def get_plugin_cache_tag(cls, path):
        """Create a checksum of the plugin module file as a tag for the cache
           database. This way the plugin cache is automatically invalidated
           everytime the module source code changes. Return value must be a
           positive integer.
        """
        with open(path, "rb") as fobj:
            return binascii.crc32(fobj.read())

    def setup(self):
        """Setup the Plugin class. This is called once for every plugin, but
           only if it will be actually used. This is mainly supposed to be used
           for importing third-party modules that may or may not be installed
           on the system.
        """

    def can_handle(self, entry):
        """Return True if this plugin can handle this particular Entry object.
           This is supposed to reduce calls to process() and minimize the
           number of needlessly cached entries.
        """
        raise NotImplementedError

    def get_entry_cache_tag(self, entry):
        """Return a value for the Entry object that if it changes causes the
           cached value to be invalidated. The default uses the modification
           time of the file. The returned value may have any type.
        """
        return entry.time

    def cache(self, entry):
        """Extract and prepare data from the Entry object and cache it. It is
           then passed on to the process() method. cache() is only called if
           use_cache is True. The data returned may have any type. Raise a
           NoData exception if the required data cannot be extracted.
        """
        # pylint:disable=unused-argument
        return None

    def process(self, entry, cached):
        """Generate key-value pairs or return a dict with the values that were
           "extracted" from the entry. The set of keys must match the
           ones from the attributes dict, but it is not necessary that all the
           keys actually have values. Raise a NoData exception or return an
           empty sequence if the plugin cannot extract the required attributes
           properly. The cached argument is only present when use_cache is True
           and contains the cached data that was prepared in the cache() method.
           If the cache() method raised a NoData exception process() will not
           be called.
        """
        raise NotImplementedError



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import shlex
import pstats
import argparse
import cProfile

from . import BaseClass, __copyright__
from .walk import Directory, FilesystemWalker
from .cache import Cache, NullCache
from .entry import StartDirectory
from .field import ExecFields, SortFields, CountFields, OutputFields, \
    ExecBatchFields
from .filter import Matcher, Excluder
from .ignore import GitIgnore
from .logger import Logger
from .parser import ParserError
from .context import Context
from .registry import Registry
from .arguments import Defaults, ArgumentsPostProcessor
from .exceptions import EX_OK, EX_PROCESS, EX_SUBPROCESS, BaseError, \
    UsageError, ProcessError, SubprocessError
from .processing import ImmediateGenerator, CollectiveGenerator, \
    ImmediateExecProcessing, CollectiveExecProcessing, \
    ImmediateConsoleProcessing, CollectiveConsoleProcessing


class _Base(BaseClass):

    def __init__(self):
        super().__init__(Context())

    def setup_processing(self):
        """Set up the required Processing object.
        """
        raise NotImplementedError

    def setup_context(self, args, warnings):
        """Initialize the central Context object which will be passed to all
           components.
        """
        for warning in warnings:
            self.logger.warning(warning)

        self.context.args = args
        self.context.setup()

        self.context.logger = Logger()
        if __debug__:
            self.context.logger.set_debug(args.debug)

    def setup_components(self):
        """Set up all remaining components like the registry, the cache and so
           on.
        """
        # pylint:disable=too-many-branches

        args = self.context.args

        # Setup the sqlite3 metadata cache.
        if args.cache is None:
            CacheClass = NullCache
        else:
            CacheClass = Cache

        # Once we're on Python >=3.8 we can use context.cache = (cache := CacheClass(context)).
        cache = CacheClass(self.context)
        self.context.cache = cache

        # Register all types and metadata plugins.
        registry = Registry(self.context)
        self.context.registry = registry

        # Replace the special value "file" in -o/--output by a list of all file
        # attributes.
        if "file" in args.output:
            pos = args.output.index("file")
            args.output.pop(pos)
            args.output = args.output[:pos] + [a + ":n" for a in registry.get_file_attributes()] + \
                    args.output[pos:]

        # Turn all options that take fields as arguments in to Fields objects.
        args.output = OutputFields(self.context, args.output)
        if args.sort is not None:
            args.sort = SortFields(self.context, args.sort)
        if args.count is not None:
            args.count = CountFields(self.context, args.count)
        if args.exec is not None:
            args.exec = ExecFields(self.context, args.exec)
        if args.exec_batch is not None:
            args.exec_batch = ExecBatchFields(self.context, args.exec_batch)

        if args.action is not None:
            if args.action == "version":
                print(__copyright__)

            elif args.action == "clean":
                self.context.cache.clean()

            else:
                manpage = getattr(registry, f"get_{args.action}_manpage")()
                manpage.show()

            raise SystemExit(EX_OK)

        elif args.help is not None:
            registry.get_plugin_manpage(args.help).show()
            raise SystemExit(EX_OK)

        try:
            # Setup the matcher from the command line arguments.
            matcher = Matcher(self.context, args.tests)
            self.context.matcher = matcher
        except ParserError as exc:
            raise UsageError(f"unable to parse tests: {exc}")

        # Setup the excluder from the command line arguments.
        excluder = Excluder(self.context, args.exclude)
        self.context.excluder = excluder

    def setup_walker(self):
        """Set up the filesystem walker processes.
        """
        walker = FilesystemWalker(self.context)
        self.context.walker = walker

        if __debug__:
            self.logger.debug("info", "Directories to search:")

        # Preload the in-queue with the path arguments.
        for path in self.context.args.directories:
            if self.context.args.absolute_path:
                path = os.path.abspath(path)

            if self.context.args.no_parent_ignore:
                # Don't use ignore files from parent directories.
                ignores = []
            else:
                # Find ignore files in parent directories.
                if self.context.args.ignore:
                    ignores = GitIgnore.from_parent_directories(self.context, os.path.dirname(path))
                else:
                    ignores = []

            if __debug__:
                self.logger.debug("info",
                        f"    {path if path != '.' else os.path.abspath(path)}")
                for ignore in ignores:
                    self.logger.debug("info",
                            f"        found ignorefile in parent: {ignore.path}")

            walker.put([Directory(StartDirectory(self.context.args, path), "", ignores)])

    def show_tests(self):
        """Show debug information about the tests to perform.
        """
        self.logger.debug("info", f"Using {self.context.processing.__class__.__name__}.")

        if not self.context.excluder.is_empty():
            self.logger.debug("info", "Exclude Sequence:")
            for line in self.context.excluder.parser.format():
                self.logger.debug("info", "  " + line)

        if not self.context.matcher.is_empty():
            self.logger.debug("info", "Test Sequence:")
            for line in self.context.matcher.parser.format():
                self.logger.debug("info", "  " + line)


class SearchNamespace(argparse.Namespace):
    """An argparser.Namespace that contains the minimum number of arguments for
       setting up the Python API.
    """
    # pylint:disable=too-many-instance-attributes

    debug = ["none"]
    profile = None
    json = None

    default_attribute = None
    default_operator = None

    depth = None

    # Options that have no relevance for the Python API but will be referenced
    # somewhere in the code. Would be nice to remove these one day.
    count = None
    exec = None
    exec_batch = None
    action = None
    help = None

    no_parent_ignore = False


class Search(_Base):
    """Query the filesystem and yield the resulting entries.
    """
    # pylint:disable=too-many-arguments,too-many-locals,dangerous-default-value,invalid-name

    def __init__(self, query=None, directories=["."], exclude=[], output=["file"],
            sort=[], reverse=False,
            hide=False, ignore=False, one_file_system=False,
            case=Defaults.case,
            si=Defaults.si,
            follow_symlinks=Defaults.follow_symlinks,
            absolute_path=False, jobs=Defaults.jobs, cache=Defaults.cache):
        super().__init__()

        if case not in Defaults.case_choices:
            raise ValueError(f"case argument must be one of {', '.join(Defaults.case_choices)}")

        if query is None:
            tests = []
        elif isinstance(query, str):
            tests = shlex.split(query)
        else:
            tests = query

        args = SearchNamespace()
        args.tests = tests
        args.directories = directories.copy()
        args.exclude = exclude.copy()
        args.output = output
        args.sort = sort
        args.reverse = reverse
        args.hide = hide
        args.ignore = ignore
        args.one_file_system = one_file_system
        args.case = case
        args.si = si
        args.follow_symlinks = follow_symlinks
        args.absolute_path = absolute_path
        args.jobs = jobs
        args.cache = cache

        processor = ArgumentsPostProcessor(args)
        processor.process()

        self.setup_context(args, [])
        self.setup_components()
        self.setup_processing()
        if __debug__:
            self.show_tests()
        self.setup_walker()

    def setup_processing(self):
        if self.context.args.sort:
            self.context.processing = CollectiveGenerator(self.context)
        else:
            self.context.processing = ImmediateGenerator(self.context)

    def __iter__(self):
        # Start processing.
        try:
            self.context.walker.start_processes()

            # Collect Entry objects from the Walker.
            self.context.processing.loop()

            # Finalize and put out the result if --sort or --exec-batch is
            # involved.
            self.context.processing.finalize()

            yield from self.context.processing

        except KeyboardInterrupt:
            # Stop all processes immediately.
            self.context.stop()
            raise

        self.context.close()
        self.context.walker.close()
        self.context.processing.close()


class Main(_Base):
    """The main entry point for the ff(1) script.
    """

    def __init__(self, args, warnings):
        super().__init__()

        self.setup_context(args, warnings)

        try:
            self.setup_components()
            self.setup_processing()
            if __debug__:
                self.show_tests()
            self.setup_walker()

        except BaseError as exc:
            self.handle_exception(exc)

    def loop(self):
        """Do all the processing and exit.
        """
        try:
            self.walk()
        except BaseError as exc:
            self.handle_exception(exc)

    def handle_exception(self, exc):
        """Print a BaseError exception using the correct formatting and exit.
        """
        if exc.traceback is not None:
            self.logger.exception(exc.message, exc.traceback, exc.exitcode)
        else:
            self.logger.error(exc.message, exc.exitcode)

    def setup_processing(self):
        args = self.context.args

        # Set up processing. There are two modes of operation: Either collect all
        # results and process them in their entirety, or process results
        # immediately as they come in.
        if args.sort or args.count or args.limit or args.exec_batch or args.json == "json":
            if args.exec or args.exec_batch:
                self.context.processing = CollectiveExecProcessing(self.context)
            else:
                self.context.processing = CollectiveConsoleProcessing(self.context)
        elif args.exec:
            self.context.processing = ImmediateExecProcessing(self.context)
        else:
            self.context.processing = ImmediateConsoleProcessing(self.context)

    def walk(self):
        """Walk through the filesystem and process the results.
        """
        try:
            if __debug__ and self.context.args.profile:
                # Start profiling. No multiprocessing is involved, Walker.loop() is
                # run in the main thread.
                profiler = cProfile.Profile()
                profiler.enable()
                self.context.walker.loop(0)
            else:
                self.context.walker.start_processes()

            # Collect Entry objects from the Walker.
            self.context.processing.loop()

            # Finalize and put out the result if --sort or --exec-batch is
            # involved.
            self.context.processing.finalize()

            if __debug__ and self.context.args.profile:
                # Stop profiling and print the statistics.
                profiler.disable()
                stats = pstats.Stats(profiler, stream=sys.stderr)
                stats.sort_stats("cumulative")
                stats.print_stats(.1)

        except KeyboardInterrupt:
            # Stop all processes immediately.
            self.context.stop()
            if __debug__:
                raise
            else:
                raise SystemExit("keyboard interrupt")

        except BrokenPipeError:
            self.context.stop()

        self.context.close()
        self.context.walker.close()
        self.context.processing.close()

        if __debug__:
            hits = self.context.cache_hits.value
            misses = self.context.cache_misses.value
            if hits or misses:
                self.logger.debug("cache", f"Cache stats: {hits} hits, {misses} misses")
            else:
                self.logger.debug("cache", "Cache was not used")

        if self.context.exitcode == EX_SUBPROCESS:
            raise SubprocessError("One or more --exec/--exec-batch commands had errors")
        elif self.context.exitcode == EX_PROCESS:
            raise ProcessError("One or more ff processes had unrecoverable errors! "\
                    "Result is probably incomplete!")



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import queue
import shutil
import threading
import traceback
import collections
import multiprocessing

from .__version__ import __version__

__copyright__= f"""ff {__version__}
Copyright (C) 2020 Lars Gustäbel <https://github.com/gustaebel/ff>
License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.

This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""


TIMEOUT = 0.01
MAX_CPU = multiprocessing.cpu_count()


class NoData(Exception):
    """Exception for when a plugin cannot process an entry.
    """

class MissingImport(Exception):
    """Exception for when a plugin cannot import a third party module.
    """
    def __init__(self, module):
        self.module = module
        super().__init__(module)


OUTPUT_WIDTH = shutil.get_terminal_size((100, 100))[0] - 2


Directory = collections.namedtuple("Directory", "start relpath ignore_paths")
Entries = collections.namedtuple("Entries", "parent entries")

class _Attribute(collections.namedtuple("Attribute", "plugin name alias")):
    """An Attribute namedtuple that contains the plugin name and the attribute
       name.
    """

    def __str__(self):
        return self.alias

    def fullname(self):
        """Return the full canonical name of the attribute.
        """
        return f"{self.plugin}.{self.name}"

    def __eq__(self, other):
        return self[:2] == other[:2]

    def __hash__(self):
        return hash(self[:2])

def Attribute(plugin, name, alias=None):
    """Factory function to create Attribute objects.
    """
    # pylint:disable=invalid-name
    if alias is None:
        alias = f"{plugin}.{name}"
    return _Attribute(plugin, name, alias)


class BaseClass:
    """This class can be used as the base for each class that depends on
       Context and access to args, registry, etc. It offers shortcuts to the
       each of these components as self.component instead of
       self.context.component without using properties. Not exactly a matter of
       life and death but nice nevertheless.
    """

    component_names = set(["args", "registry", "excluder", "matcher", "walker"])

    def __init__(self, context):
        self.context = context

    def __getattr__(self, name):
        try:
            attr = getattr(self.context, name)
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__} object has no attribute {name!r}")

        if name in self.component_names:
            setattr(self, name, attr)
        return attr


def join(part1, part2):
    """A simpler version of os.path.join() with a limited scope that is
       significantly faster.
    """
    if part1:
        return part1 + os.sep + part2
    else:
        return part2



# Put this file in ~/.ff and rename it.

from libff.plugin import *

# IMPORTANT: Do not import third-party modules here, that may not
# be installed on the system, do this in Plugin.setup().


class Foo(Plugin):
    # Give the Plugin a descriptive name that does not yet exist.

    """The "foo" plugin provides access to ...

       (This docstring will appear as the plugin's description in help texts.)
    """

    # Add some information about yourself and how to contact you.
    author = "Jane Doe"
    email = "jane@example.com"
    url = "https://example.com/foo-plugin/"

    # Here you should tell ff(1) roughly how fast this plugin is able to
    # extract data from the each entry. This is important in optimizing test
    # sequences. ff(1) will put faster tests at the beginning and slower tests
    # at the end of a test sequence to reduce the number of evaluations of
    # slower tests. There are three possible settings: FAST, NORMAL (default)
    # and SLOW:
    #
    # FAST:   Plugins that depend only on path names and inode information
    #         like stat(2) results or maybe extended attributes.
    #
    # NORMAL: Use this if you are not sure. Suitable if the plugin opens the
    #         file to read the data but does not do time-consuming
    #         processing.
    #
    # SLOW:   If the plugin has to intensively process file data (e.g.
    #         decompress it), calls external programs or libraries, or does
    #         other very slow things like web requests etc.
    speed = Speed.NORMAL

    # If use_cache is True it is possible to cache data for each entry that is
    # time-consuming to prepare. The cache() method will be called, its return
    # value will be cached and then passed to process(). Cached values are
    # discarded as soon as the associated entry's modification time changes or
    # the code of this module is altered.
    use_cache = False

    # Explicitly set the cache tag. Rarely used. If cache_tag is None it will
    # be set to the return value of get_plugin_cache_tag(). See there for a
    # more in-depth explanation.
    cache_tag = None

    # Here you tell ff(1) which attributes your plugin will provide.
    # `attributes` is a list of 3-element-tuples: (name, type, description).
    # `type` must be a subclass of Type. A Type's job is for example to
    # describe which operators the value supports, and to format the value
    # for output or to a human readable form. You can create your own Types if
    # needed, see libff/type.py for details.
    #
    # Available Types are:
    #
    #    String             a generic string
    #    ListOfStrings      a list of strings
    #    Number             a number greater or equal to 0
    #    Size               a size in bytes
    #    Time               a point in time in seconds
    #    Duration           a duration in seconds
    #    Boolean            a true or false value
    #
    attributes = [
        ("foo", String, "The name of the file."),
        ("bar", Size, "The size of the file.")
    ]

    @classmethod
    def get_plugin_cache_tag(cls, path):
        # Optional. Is only used if use_cache = True. Rarely used.
        #
        # Return a "tag" that should stay the same as long as the format of the
        # plugin's cached data stays the same. Otherwise, all cached values for
        # this plugin are discarded and recached.
        # The default scheme returns a checksum of the plugins source code, so
        # the cache is invalidated automatically everytime the plugin code is
        # changed.
        # If you need more control over the cache you can implement your own
        # scheme, e.g. return a number that you increment manually everytime
        # you change the data format.

    @classmethod
    def setup(cls):
        # Optional.
        #
        # This classmethod is called once before the Plugin is about to be
        # used. It is supposed to be used for setup code, mainly for importing
        # modules that may or may not be installed.
        #
        #   @classmethod
        #   def setup(cls):
        #       global module
        #       import module

    def can_handle(self, entry):
        # Required.
        #
        # Return True if this plugin can handle this type of file. There are
        # many ways possible for you to do this: based on an attribute of the
        # Entry object (e.g. file type, extension, mime type etc.) or by
        # reading a magic number from the file.
        # This check is supposed to reduce the number that the cache() and
        # process() methods will be called and will also reduce the load on the
        # cache and the number of needlessly cached entries.
        #
        #   def can_handle(self, entry):
        #       return entry.is_file()

    def get_entry_cache_tag(self, entry):
        # Optional. Is only called if use_cache = True. Rarely used.
        #
        # Return a "tag" that indicates the validity of the cached value for
        # this entry. Once it changes the cached value will be discarded and
        # recached. The default method returns the file's modification time
        # which is what you want in most cases.

    def cache(self, entry):
        # Optional. Is only called if use_cache = True.
        #
        # Extract and prepare information from the entry and put it in the
        # cache. After the value has been cached this method will not be called
        # again unless the result of cache_tag(entry) changes. The return value
        # of cache() will be cached and then passed to the process() method. It
        # may have any type. If the file information cannot be extracted, raise
        # a NoData exception.

    def process(self, entry, cached):
        # Required.
        #
        # Extract information from the entry to produce values for the
        # attributes described above. `entry` is a an Entry object from
        # libff/entry.py, see there for attributes and properties. If use_cache
        # is True, `cached` is the cached return value from the cache() method
        # or None if there was a NoData exception raised.
        # On success, either return a dictionary with the values or yield
        # key-value pairs. On failure, i.e. if the file is invalid or the
        # required information cannot be extracted, raise a NoData exception or
        # return an empty dictionary.
        yield "foo", entry.name
        yield "bar", entry.size



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import re
import sys
import shlex
import argparse

from . import MAX_CPU
from .exceptions import EX_OK, UsageError


class Defaults:
    """Provide the defaults for both the ArgumentParser as well as the API.
    """

    case = "smart"
    case_choices = ["smart", "ignore", "sensitive"]
    default_attribute = "name"
    default_operator = "~"
    follow_symlinks = False
    jobs = MAX_CPU
    cache = os.path.expanduser("~/.cache/ff.db")
    si = False


def type_jobs(string):
    """Parse the number of jobs using a default of MAX_CPU if it is zero.
    """
    num = type_number(string)
    if num == 0:
        num = MAX_CPU
    return max(min(num, MAX_CPU), 1)


def type_number(string):
    """Parse a number greater than zero.
    """
    num = int(string)
    if num < 0:
        raise ValueError("number must be greater than zero")
    return num


def type_list(string):
    """Parse a comma separated string into a list.
    """
    return [s.strip() for s in string.split(",")]


regex_range = re.compile(r"^(?:(?P<single>\d+)|(?P<start>\d*)-(?P<stop>\d*))$")

def type_ranges(string):
    """Parse a comma separated list of ranges.
    """
    segments = []
    for range_ in type_list(string):
        match = regex_range.match(range_)
        if match is None:
            raise ValueError("invalid range")

        if match.group("single") is not None:
            number = int(match.group("single"))
            segments.append((number, number))

        else:
            start = match.group("start")
            if start:
                start = int(start)
            else:
                start = 0

            stop = match.group("stop")
            if stop:
                stop = int(stop)
                if stop < start:
                    stop, start = start, stop
            else:
                stop = None

            segments.append((start, stop))

    return segments


class HelpFormatter(argparse.HelpFormatter):
    """HelpFormatter subclass that does not show arguments for short but only
       for long options and shows the argument for REMAINDER options instead of
       three dots.
    """

    def __init__(self, prog, indent_increment=2, max_help_position=27, width=None):
        super().__init__(prog, indent_increment, max_help_position, width)

    def _format_action_invocation(self, action):
        # Don't show the argument for short options. This looks cleaner.
        if not action.option_strings:
            return super()._format_action_invocation(action)
        else:
            parts = []
            if action.nargs == 0:
                parts.extend(action.option_strings)
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for i, option_string in enumerate(action.option_strings):
                    if i == 0:
                        parts.append(option_string)
                    else:
                        parts.append('%s %s' % (option_string, args_string))
            return ', '.join(parts)

    def _format_args(self, action, default_metavar):
        # Instead of three dots ... show the name of the argument for REMAINDER
        # arguments.
        result = super()._format_args(action, default_metavar)
        if action.nargs == argparse.REMAINDER:
            return self._metavar_formatter(action, default_metavar)(1)[0]
        else:
            return result


def create_parser(formatter_class=HelpFormatter):
    """Create the argparse.ArgumentParser object.
    """
    # pylint:disable=too-many-statements
    parser = argparse.ArgumentParser(prog="ff", formatter_class=formatter_class, add_help=False,
            usage="%(prog)s [<options>] [<test/directory> ... | [-D] <directory> ...]")

    group = parser.add_argument_group("Global options")

    group.add_argument("tests", nargs="*", metavar="<test/directory>",
            help="A test expression for matching files or a directory to search.")
    if __debug__:
        group.add_argument("--profile", action="store_true", default=False,
                help="Do a profiling run on the given arguments and suppress the output.")
        group.add_argument("--debug", type=type_list, default=None,
                help="Show only debug messages of certain categories, default is to show all.")
    group.add_argument("--cache", default=Defaults.cache,
            help="Location of the metadata cache (default: %(default)s).")
    group.add_argument("--no-cache", action="store_const", dest="cache", const=None,
            help="Do not use the metadata cache.")
    group.add_argument("--clean-cache", action="store_const", const="clean", dest="action",
            help="Remove stale entries from the metadata cache.")
    group.add_argument("-j", "--jobs", type=type_jobs, default=Defaults.jobs, metavar="<num>",
            help="Set number of processes to use for searching and executing "\
                 "(default: the number of CPU cores).")
    group.add_argument("-D", "--directories", nargs="+", default=[], metavar="<path>",
            help="Search entries in these paths (default is current directory).")

    group = parser.add_argument_group("Commands")
    group.add_argument("-h", "--help", nargs="?", const="all", metavar="<plugin>",
            help="Show this help message or the help message for a particular plugin.")
    group.add_argument("--version", action="store_const", const="version", dest="action",
            default=None, help="Show program's version number and exit.")
    group.add_argument("--help-full", action="store_const", const="full", dest="action",
            help="Show a full help in man page format.")
    group.add_argument("--help-attributes", action="store_const", const="attributes", dest="action",
            help="Show a list of available attributes to use for searching, sorting and output.")
    group.add_argument("--help-plugins", action="store_const", const="plugins", dest="action",
            help="Show a list of available plugins.")
    group.add_argument("--help-types", action="store_const", const="types", dest="action",
            help="Show a list of available types.")

    group = parser.add_argument_group("Search options")
    group.add_argument("-H", "--hide", action="store_true",  default=False,
            help="Do not show hidden files and directories.")
    group.add_argument("-I", "--ignore", action="store_true", default=False,
            help="Do not show files that are excluded by patterns from .(git|fd|ff)ignore files.")
    group.add_argument("-d", "--depth", type=type_ranges, default=None, metavar="<range>",
            help="Show only files that are located at a certain depth level of the directory "\
                 "tree that is within the given <range>. A <range> is a string of the form "\
                 "'<start>-<stop>'. <start> and <stop> are optional and may be omitted. "\
                 "<range> may also be a single number. It is possible to specify multiple "\
                 "ranges separated by comma.")
    group.add_argument("--no-parent-ignore", action="store_true", default=False,
            help="Do not read patterns from ignore files from parent directories.")
    group.add_argument("-e", "--exclude", action="append", default=[], metavar="<test>",
            help="Exclude entries that match the given test.")
    group.add_argument("-c", "--case", choices=Defaults.case_choices, dest="case",
            default=Defaults.case, metavar="<mode>",
            help="How to treat the case of text attributes (smart, ignore or sensitive).")
    group.add_argument("-L", "--follow", action="store_true", dest="follow_symlinks",
            default=Defaults.follow_symlinks, help="Follow symbolic links.")
    group.add_argument("--one-file-system", "--mount", "--xdev", action="store_true", default=False,
            help="Do not descend into different file systems.")

    group = parser.add_argument_group("Simple pattern options")
    group.add_argument("-g", "--glob", action="store_const", const="%", dest="default_operator",
            default=Defaults.default_operator, help="Treat the pattern as a glob(7) pattern.")
    group.add_argument("-r", "--regex", action="store_const", const="~", dest="default_operator",
            help="Perform a regular-expression based search (default).")
    group.add_argument("-F", "--fixed-strings", action="store_const", const=":",
            dest="default_operator", help="Treat the pattern as a literal string.")
    group.add_argument("-p", "--full-path", action="store_const", const="path",
            dest="default_attribute", default=Defaults.default_attribute,
            help="Search full path (default: basename only).")

    group = parser.add_argument_group("Output options")
    group.add_argument("-x", "--exec", nargs=argparse.REMAINDER, metavar="<cmd>",
            help="Execute a command for each search result.")
    group.add_argument("-X", "--exec-batch", nargs=argparse.REMAINDER, metavar="<cmd>",
            help="Execute a command with all search results at once.")
    group.add_argument("-C", "--color", choices=["auto", "never", "always"],
            default="never" if "NO_COLOR" in os.environ else "auto",
            metavar="<when>", help="When to use colors: never, *auto*, always.")
    group.add_argument("-a", "--absolute-path", action="store_true", default=False,
            help="Show absolute instead of relative paths.")
    group.add_argument("-0", "--print0", action="store_const", const="\0", dest="newline",
            default="\n", help="Separate results by the null character.")
    group.add_argument("-v", "--verbose", action="store_const", dest="output", default=["path"],
            const=["mode:h", "links", "user:h", "group:h", "size:5h", "time:h", "path:h"],
            help="Produce output similar to `ls -l`.")
    group.add_argument("-S", "--sort", nargs="?", type=type_list, const="file.path", default=None,
            metavar="<attribute-list>", help="Sort entries by path or any other attribute.")
    group.add_argument("-R", "--reverse", action="store_true", default=False,
            help="Reverse the sort order.")
    group.add_argument("--count", nargs="?", type=type_list, const="file.size:h,file.type",
            default=None, metavar="<attribute-list>",
            help="Count the attributes from <attribute-list> and print statistics, "\
                 "instead of the result, the default is to count the total size and "\
                 "the file types of the entries found. Add --json for JSON output.")
    group.add_argument("-l", "--limit", action="store", type=type_number, default=None, metavar="N",
            help="Limit output to at most N entries.")
    group.add_argument("-1", action="store_const", const=1, dest="limit",
            help="Print only the first entry and exit immediately.")
    group.add_argument("-o", "--output", type=type_list, metavar="<attribute-list>",
            help="Print each entry by using a template of comma-separated attributes. "\
                 "The special value 'file' stands for all file attributes.")
    group.add_argument("--sep", default=" ", metavar="<string>", dest="separator",
            help="Separate each attribute of --output with <string>, default is a single space.")
    group.add_argument("--all", action="store_true", default=False,
            help="Show all entries including the ones with missing attribute values.")
    group.add_argument("--json", action="store_const", const="json", dest="json", default=None,
            help="Print attributes as one big json object to stdout.")
    group.add_argument("--jsonl", action="store_const", const="jsonl", dest="json",
            help="Print attributes as jsonl (one json object per line) to stdout.")
    group.add_argument("--si", action="store_true", default=Defaults.si,
            help="Parse and print file sizes in units of 1K=1000 bytes instead of 1K=1024 bytes.")

    return parser


def collect_arguments():
    """Join arguments for the ArgumentParser from the FF_OPTIONS environment
       variable with the ones from the command line.
    """
    ff_options = os.environ.get("FF_OPTIONS")
    if ff_options:
        argv = shlex.split(ff_options)
    else:
        argv = []
    argv += sys.argv[1:]
    return argv


class ArgumentsPostProcessor:
    """Check and postprocess the arguments in an argparse.Namespace.
    """

    def __init__(self, args):
        self.args = args

    def process(self):
        """Check and postprocess the arguments.
        """
        self.process_directories()
        self.process_arguments()
        self.check_for_errors()
        return self.collect_warnings()

    def check_for_errors(self):
        """Check some of the arguments for semantic conflicts with other arguments.
        """
        if __debug__:
            if self.args.profile and (self.args.exec or self.args.exec_batch):
                raise UsageError("You cannot use --exec or --exec-batch together with --profile!")

        if self.args.count and (self.args.exec or self.args.exec_batch):
            raise UsageError("You cannot use --exec or --exec-batch together with --count!")

        if self.args.count and self.args.limit is not None:
            raise UsageError("You cannot use --limit together with --count!")

    def collect_warnings(self):
        """Check some of the arguments for semantic conflicts with other arguments.
        """
        warnings = []

        if self.args.sort and self.args.exec and self.args.jobs != 1:
            warnings.append("Using both --sort and --exec makes no sense unless you set --jobs=1!")

        if self.args.output != ["path"] and (self.args.exec or self.args.exec_batch):
            self.args.output = ["path"]
            warnings.append(
                    "Switching off --output, it has no effect with --exec and --exec-batch.")

        if self.args.count:
            if self.args.sort:
                self.args.sort = None
                warnings.append("Switching off --sort, it has no effect with --count.")

            if self.args.output != ["path"]:
                self.args.output = ["path"]
                warnings.append("Switching off --output, it has no effect with --count.")

        return warnings

    def process_directories(self):
        """Arrange directory arguments and check them for validity.
        """
        if not self.args.directories:
            # Check which arguments are existing directories and append them to
            # self.args.directories. We allow directory arguments only at the start or
            # the end of the list of tests.
            for tests in (list(self.args.tests), reversed(self.args.tests)):
                for test in tests:
                    if os.sep in test and os.path.isdir(test):
                        self.args.tests.remove(test)
                        directory = os.path.normpath(test)
                        self.args.directories.append(directory)
                    else:
                        break

        # Default to the current directory if no directory arguments are specified
        # or detected.
        if not self.args.directories:
            self.args.directories = ["."]

        # Check if directory arguments are sub-directories of one another.
        for directory in sorted(self.args.directories, reverse=True):
            for subdir in self.args.directories:
                if subdir == directory:
                    continue
                if os.path.commonpath([directory, subdir]) == subdir:
                    raise UsageError(f"{directory!r} is a sub-directory of {subdir!r}")

    def process_arguments(self):
        """Check existing arguments and arrange them in specific ways.
        """
        if self.args.json:
            self.args.color = "never"
            self.args.absolute_path = True

        if self.args.ignore:
            self.args.exclude.append("ignored=yes")

        if self.args.one_file_system:
            self.args.exclude.append("samedev=no")

        if self.args.hide:
            self.args.exclude.append("hide=yes")

        if self.args.depth:
            # Look for an upper limit in the given ranges and exclude
            # everything above it, so we go only as deep into the tree as
            # necessary.
            if not [stop for start, stop in self.args.depth if stop is None]:
                stop = max(stop for start, stop in self.args.depth)
                self.args.exclude.append(f"depth+{stop}")

            # Construct a set of tests for all the given ranges and embed the
            # existing set of tests in it.
            tests = []
            tests.append("{{")

            for i, (start, stop) in enumerate(self.args.depth):
                if start == stop:
                    tests.append(f"depth={start}")
                else:
                    tests.append(f"depth+={start}")
                    if stop is not None:
                        tests.append(f"depth-={stop}")

                if i < len(self.args.depth) - 1:
                    tests.append("OR")

            tests.append("}}")

            if self.args.tests:
                tests.append("{{")
                tests += self.args.tests
                tests.append("}}")

            self.args.tests = tests


def parse_arguments():
    """Parse the arguments from the command line, check for conflicts and
       postprocess them for later use.
    """
    parser = create_parser()

    argv = collect_arguments()
    args = parser.parse_args(argv)

    if args.help == "all":
        # Help on plugins is taken care of in libff/search.py.
        parser.print_help()
        raise SystemExit(EX_OK)

    processor = ArgumentsPostProcessor(args)
    warnings = processor.process()

    return args, warnings



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import re
import collections

from . import Attribute
from .type import Count
from .exceptions import UsageError

Field = collections.namedtuple("Field", "attribute type width modifier")


class Fields(list):
    """A class that stores a list of attributes as fields.
    """

    regex = re.compile(r"^((?:[a-zA-Z][a-zA-Z0-9_]+\.)?[a-zA-Z][a-zA-Z0-9_]+?)"\
            r"(?::(-?\d+)?(h|o|x|n)?)?$")

    def __init__(self, context, argument):
        super().__init__()

        self.context = context
        self.args = self.context.args
        self.registry = self.context.registry

        self.store(argument)

    def make_field(self, string):
        """Create a Field from an attribute string.
        """
        match = self.regex.match(string.strip())
        if match is None:
            raise UsageError(f"Invalid attribute {string!r}")
        name, width, modifier = match.groups()
        attribute = self.registry.setup_attribute(name)
        type_cls = self.registry.get_attribute_type(attribute)
        return Field(attribute, type_cls, width, modifier)

    def store(self, argument):
        """Convert argument to a list of Field objects and store them.
        """
        raise NotImplementedError


class OutputFields(Fields):
    """Store a list of fields that help with formatting output.
    """

    def store(self, argument):
        for string in argument:
            self.append(self.make_field(string))

    def to_dict(self, entry):
        """Create a dictionary from the list of fields.
        """
        record = {}
        for field in self:
            try:
                value, type_cls = \
                        self.context.registry.get_attribute_and_type(entry, field.attribute)
            except KeyError:
                value = None
            else:
                if field.modifier == "h":
                    value = type_cls.output(self.args, field.modifier, value)

            record[str(field.attribute)] = value

        return record


class CountFields(OutputFields):
    """Store a list of fields that help with formatting statistics output.
    """

    def store(self, argument):
        super().store(argument)

        for field in self:
            if field.type.count is Count.UNCOUNTABLE:
                raise UsageError(f"Attribute {field.attribute} is not suited for --count!")


class SortFields(OutputFields):
    """Store a list of fields that help with sorting Entry objects.
    """

    def render(self, entry):
        """Create a sort key from an Entry object.
        """
        output = []
        for field in self:
            try:
                value = self.registry.get_attribute(entry, field.attribute)
            except KeyError:
                output.append(field.type.sort_none)
            else:
                value = field.type.sort_key(value)
                output.append(value)
        return tuple(output)


class ExecFields(Fields):
    """Store a list of fields that help with calling subprocesses.
    """

    def store(self, argument):
        for string in argument:
            if string == "{}":
                field = self.make_field("file.path")
            elif string == "{/}":
                field = self.make_field("file.name")
            elif string == "{//}":
                field = self.make_field("file.dir")
            elif string == "{.}":
                field = self.make_field("file.pathx")
            elif string == "{/.}":
                field = self.make_field("file.namex")
            elif string == "{..}":
                field = self.make_field("file.ext")
            elif string.startswith("{") and string.endswith("}"):
                field = self.make_field(string[1:-1])
            else:
                field = string

            self.append(field)

        # Append path if there is no placeholder in the argument list.
        if not any(isinstance(field, Field) for field in self):
            attribute = Attribute("file", "path")
            type_cls = self.registry.get_attribute_type(attribute)
            self.append(Field(attribute, type_cls, None, None))

    def render(self, entry):
        """Create a list of arguments from an Entry object for calling a
           subprocess.
        """
        return self.render_fields(self, [entry], ignore_missing=False)

    def render_fields(self, fields, entries, ignore_missing=True):
        """Return a list of arguments with all placeholders replaced.
        """
        output = []
        for field in fields:
            if isinstance(field, Field):
                for entry in entries:
                    try:
                        value = self.registry.get_attribute(entry, field.attribute)
                    except KeyError:
                        if self.args.all or field.modifier == "n":
                            output.append("")
                        elif not ignore_missing:
                            raise
                    else:
                        output.append(field.type.output(self.args, field.modifier, value))
            else:
                # Fields that do not contain a placeholder will be left
                # unchanged.
                output.append(field)
        return output


class ExecBatchFields(ExecFields):
    """Store a list of fields that help with calling one subprocess with
       arguments from multiple Entry objects.
    """

    # pylint:disable=arguments-differ
    def render(self, entries):
        """Create a list of arguments from multiple Entry objects for calling a
           subprocess.
        """
        return [self[0]] + self.render_fields(self[1:], entries, ignore_missing=True)
